/*
   _===_ _           _                  ______ _     _             _ _ _
  |  ___| |         | |                 | ___ (_)   | |           (_) | |
  | |__ | | ___  ___| |_ _ __ ___ ______| |_/ /_  __| | ___  _   _ _| | | ___ _   _ _ __
  |  __|| |/ _ \/ __| __| '__/ _ \______| ___ \ |/ _` |/ _ \| | | | | | |/ _ \ | | | '__|
  | |___| |  __/ (__| |_| | | (_) |     | |_/ / | (_| | (_) | |_| | | | |  __/ |_| | |
  \____/|_|\___|\___|\__|_|  \___/      \____/|_|\__,_|\___/ \__,_|_|_|_|\___|\__,_|_|

  Contrôleur de commutateur Coaxial 8:1 Ameritron (ou similaire), Version 1.
  Utilise un Raspberri Pi Pico. Développé sous Arduino IDE.
  Électro-Bidouilleur, Juin 2023. https://www.youtube.com/Électro-Bidouilleur
  Licence de droits: Attribution 4.0 International (CC BY 4.0)  Lisez https://creativecommons.org/licenses/by/4.0/deed.fr
*/

// DIRECTIVES AU COMPILATEUR
#include "Adafruit_ILI9341.h"
#include "Adafruit_GFX.h"
#include <XPT2046_Touchscreen.h>
#include <EEPROM.h>

// Voici les connexions sur un Rasprerry Pi Pico. Les numéros indiqués correspondent aux appellations GPxx du brochage.
#define TFT_DC 21
#define TFT_CS 17
#define TFT_MOSI 19
#define TFT_CLK 18
#define TFT_RST 20
#define TFT_MISO 16
#define TCS_PIN  13
#define TIRQ_PIN 14
#define control_eclair_affich 5
#define Relai_A 10
#define Relai_B 11
#define Relai_C 12
#define PTT_In 2
#define Vibreur 22

// CONSTANTES
const String Data_Antennes[8][4] {{"1 Dummy Load",       "TX-RX",   "0",  "0"},               // Le texte qui apparaîtra sur les étiquettes pour chaque antenne. Maximum 16 caractères
                                  {"2 20/15/10 Yagi",    "TX-RX",   "11", "40"},
                                  {"3 40/80 H-Sloper",   "TX-RX",   "1",  "11"},              //"TX-RX" = Antenne utilisable en TX
                                  {"4 6m Moxon",         "TX-RX",   "40", "65"},              //"RX" = Antenne utilisable seulement en Rx
                                  {"5 Loop-On-Ground",   "RX",      "0",  "0"},               // Les deux nombres: fréquences minimum (>=) et maximum (<).
                                  {"6 Mag Loop",         "RX",      "0",  "0"},               //   0 et 0 rendent une antenne manuelle, i.e. jamais sélectionnée automatiquement.
                                  {"7 Long Wire",        "TX-RX",   "0",  "0"},
                                  {"8    -----      ",   "RX",      "0",  "0"}};

// VARIABLES GLOBALES
unsigned long int debut_compte_affich_millis = 0;               // Valeurs initiales de la fonction millis() pour calculer des délais.
unsigned long int debut_compte_clignot_millis = 0;              //      "
unsigned long int debut_rafraich_pause_millis = 0;              //      "
unsigned int antenne_Tx = 0;                                    // Antenne Tx sélectionnée, de 0 à 7
unsigned int antenne_Tx_old;                                    // Antenne Tx préalablement sélectionnée, de 0 à 7
unsigned int antenne_Rx = 0;                                    // Antenne Rx sélectionnée, de 0 à 7
double moy_px = 0;                                              // Variables de position tactile
double moy_py = 0;                                              //      "
int moy_compte = 0;                                             //      "
bool ptt_actif = false;                                         // État du PTT (Tx=1, Rx=0)
bool alarme_antenne_rx = false;                                 // Fanion d'alarme de Tx sur une antenne Rx seulement
bool clignot_on;                                                // Fanion de clignotement de l'étiquette Tx
unsigned int caract;                                            // Caractère reçu sur le port CI-V
unsigned int rx_data[15];                                       // Chaîne de caractères reçus sur le port CI-V
unsigned int byte_pos = 0;                                      // Position du caractère lu sur la chaîne de caractères
unsigned int frequence;                                         // Fréquence du transceiver, exprimée en MHz et arrondie
int old_frequence = -1;                                         // Fréquence préalable du transceiver
bool nouv_frequence_dispo = true;                               // Fanion d'avertissement qu'une nouvelle fréquence est disponible.

// INSTANCIATION DES CLASSES
// Utilisation du port SPI intégré (port matériel) du R-Pi Pico pour l'affichage. Il faut spécifier la position des broches CS, DC et RST pour l'affichage.
Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_RST);
// Implémentation de la lecture tactile en utilisant la broche IRQ
XPT2046_Touchscreen ts(TCS_PIN, TIRQ_PIN);  // Param 2 - Touch IRQ Pin - interrupt enabled polling

// Fonction produisant un court bip sonore sur le vibreur magnétique
void vibreur_bip()
{
  tone(Vibreur, 2400);                                            // 2400 Hz, la fréquence suggérée du vibreur magnétique
  debut_rafraich_pause_millis = millis();
  while (millis() <= (debut_rafraich_pause_millis + 30));         // Une durée de 30 millisecondes
  noTone(Vibreur);
}

// Fonction d'initialisation du premier coeur du Pico, style Arduino, exécutée une seule fois à la mise sous tension
void setup()
{
  pinMode(TCS_PIN, OUTPUT);                               // S'assurer que les broches CS sont au niveau haut pour commencer
  digitalWrite(TCS_PIN, HIGH);                            //    "
  pinMode(TFT_CS, OUTPUT);                                //    "
  digitalWrite(TFT_CS, HIGH);                             //    "
  digitalWrite(Relai_A, LOW);                             //    "
  pinMode(Relai_A, OUTPUT);                               // Les broches de relais au niveau bas pour commencer
  digitalWrite(Relai_B, LOW);                             //    "
  pinMode(Relai_B, OUTPUT);                               //    "
  digitalWrite(Relai_C, LOW);                             //    "
  pinMode(Relai_C, OUTPUT);                               //    "
  pinMode(PTT_In, INPUT);                                 // La broche PTT_In initalisée en entrée
  pinMode(control_eclair_affich, OUTPUT);                 // Initialiser la broche de contrôle d'éclairage de l'affichage comme sortie et allumer l'éclairage
  digitalWrite(control_eclair_affich, HIGH);              //    "

  tft.begin(24000000);                                    // Démarrer l'instance d'écran
  tft.setRotation(3);                                     // Faire pivoter l'affichage de 270 degrés
  ts.begin();                                             // Démarrer l'instance de détection tactile
  EEPROM.begin(256);                                      // Démarrer l'instance d'accès EEPROM (émulation dans la Flash)

  while (!Serial && (millis() <= 1000));                  // Une pause d'au moins 1 seconde

  // Les antennes sauvegardées en EEPROM sont restaurées par défaut.
  antenne_Tx = EEPROM.read(0);                            // Lecture de la valeur sauvegardée de l'antenne TX
  antenne_Rx = EEPROM.read(1);                            // Lecture de la valeur sauvegardée de l'antenne RX
  if (antenne_Tx > 7)                                     // Couvre la première mise sous tension avec la mémoire EEPROM vide
  {
    antenne_Tx = 0;                                       // Initialiser les deux antennes à 0 (antenne 1)
    antenne_Rx = 0;
    EEPROM.write(0, antenne_Tx);                          // Sauvegarder les valeurs d'antennes dans la EEPROM
    EEPROM.write(1, antenne_Rx);
    EEPROM.commit();
  }
  tft.fillScreen(ILI9341_BLACK);                          // Effacer l'écran au noir

  // Dessin des bannières d'antennes et addition du texte
  tft.setTextSize(2);
  tft.setTextColor(ILI9341_WHITE);                        // Couleur du texte
  for (int i = 0; i <= 7; i++)                            // 8 bannières d'antenne
  {
    tft.fillRect(16, 10 + i * 28, 200, 23, ILI9341_BLUE); // Dessin des bannières bleues
    tft.setCursor(20, 15 + i * 28);                       // Écriture du texte des antennes
    tft.print(Data_Antennes[i][0]);
  }
}

// Fonction d'initialisation du deuxième coeur du Pico, style Arduino, exécutée une seule fois à la mise sous tension
void setup1()
{
  //  Serial.begin(9600);                                   // Port debug USB. Activer seulement pour déboguer
  Serial1.begin(4800);                                    // Port série connecté au transceiver (protocole CI-V)

  // Envoi d'une requête initiale de fréquence. Si pas de réponse obtenue, les antennes en EEPROM seront applicables.
  const byte tx_data[] = {0xFE, 0xFE, 0x94, 0xE0, 0x03, 0xFD}; // La syntaxe du message de requête de fréquence
  Serial1.write(tx_data, sizeof(tx_data));                // Envoi du message
}

// La boucle principale, style Arduino. Elle tourne de façon perpétuelle sur le premier coeur du Pico.
void loop()
{
  // Lecture du PTT et gestion des antennes "RX"
  if (ptt_actif != !digitalRead(PTT_In))                            // Niveau bas actif sur la broche PTT_In?
  {
    ptt_actif = !digitalRead(PTT_In);                               // Oui, changer l'état de la variable PTT
    if (ptt_actif && (Data_Antennes[antenne_Tx][1] == "RX"))        // Antenne ne permettant pas d'émettre (Tx)?
    {
      antenne_Tx_old = antenne_Tx;                                  // Sauvegarder l'antenne préalable
      antenne_Tx = 0;                                               // Transférer vers l'antenne 0 (charge fictive)
      alarme_antenne_rx = true;                                     // Signaler une alarme d'antenne
    }
    else if (alarme_antenne_rx)                                     // Passage en réception avec alarme Rx active
    {
      antenne_Tx = antenne_Tx_old;                                  // Remettre l'ancienne antenne Rx
      alarme_antenne_rx = false;                                    // Désactiver l'alarme
    }
  }

  // Mise à jour des relais d'antennes
  if (ptt_actif)
  {
    digitalWrite(Relai_A, bitRead(antenne_Tx, 0));                 // Sélectionne les bons relais en Tx en fonction du choix de l'usager
    digitalWrite(Relai_B, bitRead(antenne_Tx, 1));
    digitalWrite(Relai_C, bitRead(antenne_Tx, 2));
  }
  else
  {
    digitalWrite(Relai_A, bitRead(antenne_Rx, 0));                 // Sélectionne les bons relais en Rx en fonction du choix de l'usager
    digitalWrite(Relai_B, bitRead(antenne_Rx, 1));
    digitalWrite(Relai_C, bitRead(antenne_Rx, 2));
  }

  // Rafraîchissement des carrés TX et RX
  for (int i = 0; i <= 7; i++)
  {
    if (ptt_actif)                                                  // En PTT actif, les cases TX rouges sont affichées et les cases Rx sont en gris
    {
      tft.drawRect(226, 10 + i * 28, 36, 23, ILI9341_RED);          // Dessin des boutons TX
      tft.drawRect(270, 10 + i * 28, 36, 23, 0x4108);               // Gris très foncé. RGB=5bits/6bits/5bits
    }
    else                                                            // En PTT inactif, les cases Rx vertes sont affichées et les cases Tx sont en gris
    {
      tft.drawRect(226, 10 + i * 28, 36, 23, 0x4108);               // Gris très foncé. RGB=5bits/6bits/5bits// Dessin des boutons RX
      tft.drawRect(270, 10 + i * 28, 36, 23, ILI9341_GREEN);        // Dessin des boutons RX
    }
  }

  // Affichage des étiquettes "TX" et "RX" des relais
  for (int i = 0; i <= 7; i++)                                      // Parcours de toute les antennes pour identifier laquelle est sélectionnée
  {
    // Gestion de l'étiquette TX
    tft.setTextColor(ILI9341_RED);                                  // Le "TX" sera en rouge
    tft.setCursor(232, 15 + i * 28);
    if (i == antenne_Tx)                                            // L'antenne sélectionnée pour le Tx?
    {
      if ((Data_Antennes[i][1] == "TX-RX")) tft.print("TX");        // Écrire "TX" dans la case correspondant à l'antenne sélectionée si elle est permise en Tx
      else tft.print("--");                                         // Sinon écrire "--" Pour une antenne de réception uniquement
    }
    else tft.fillRect(232, 15 + i * 28, 25, 16, ILI9341_BLACK);     // Effacer les autres cases.
    if ((alarme_antenne_rx) && (!clignot_on)) tft.fillRect(232, 15 + i * 28, 25, 16, ILI9341_BLACK);  // Produit le clignotement du "TX" durant l'alarme

    // Gestion de l'étiquette RX
    tft.setTextColor(ILI9341_GREEN);                                // Le "RX" sera en vert
    tft.setCursor(277, 15 + i * 28);
    if (i == antenne_Rx) tft.print("RX");                           // Écrire "RX" dans la case correspondant à l'antenne sélectionée
    else tft.fillRect(277, 15 + i * 28, 25, 16, ILI9341_BLACK);     // Effacer les autres cases.
  }
  // Traitement des actions tactiles
  if (ts.touched())                                                 // Affichage touché?
  { // Oui
    if (!digitalRead(control_eclair_affich))                        // Affichage éteint?
    { // Oui
      debut_compte_affich_millis = millis();                        // Réinitialiser la minuterie d'éteignement de l'affichage
      digitalWrite(control_eclair_affich, HIGH);                    // Allumer l'affichage
      while (ts.touched());                                         // Attendre le relâchement de la touche
    }
    while (ts.touched())                                            // Boucler tant qu'il y a détection de touche
    { // Présentement touché
      TS_Point p = ts.getPoint();                                   // Obtenir les coordonnées de touche
      moy_px += p.x;                                                // Calcul de la moyenne des coordonnées (augmente la précision)
      moy_py += p.y;                                                //     "
      moy_compte++;                                                 //     "
    }
    debut_rafraich_pause_millis = millis();
    while (millis() <= (debut_rafraich_pause_millis + 100));        // Une pause de 100 millisecondes
    moy_px = round(float(moy_px) / moy_compte);                     // Calcul de la moyenne des coordonnées (augmente la précision)
    moy_py = round(float(moy_py) / moy_compte);

    // Décodage du bouton TX sélectionné en fonction des coordonnées
    if ((moy_px > 2800 && moy_px <= 3200) && (!ptt_actif))          // Décodage de la colonne de boutons TX
    {
      if      (moy_py > 300 && moy_py <= 675)   antenne_Tx = 0;     // Décodage de la ligne du bouton sélectionné
      else if (moy_py > 676 && moy_py <= 1125)  antenne_Tx = 1;     //     "
      else if (moy_py > 1126 && moy_py <= 1575) antenne_Tx = 2;
      else if (moy_py > 1576 && moy_py <= 2025) antenne_Tx = 3;
      else if (moy_py > 2026 && moy_py <= 2475) antenne_Tx = 4;
      else if (moy_py > 2476 && moy_py <= 2925) antenne_Tx = 5;
      else if (moy_py > 2926 && moy_py <= 3375) antenne_Tx = 6;
      else if (moy_py > 3376 && moy_py <= 3750) antenne_Tx = 7;
      antenne_Rx = antenne_Tx;                                      // L'antenne RX doit suivre l'antenne TX lorsqu'une sélection d'antenne TX est faite
      EEPROM.write(0, antenne_Tx);                                  // Sauvegerder en EEPROM l'antenne TX sélectionnée
      EEPROM.write(1, antenne_Rx);                                  // Sauvegerder en EEPROM l'antenne RX sélectionnée
      EEPROM.commit();                                              // Sauvegarde finale ici
      vibreur_bip();
    }
    // Décodage du bouton RX sélectionné en fonction des coordonnées
    else if (moy_px > 3300 && moy_px <= 3700)                       // Décodage de la colonne de boutons RX
    {
      if      (moy_py > 300 && moy_py <= 675)  antenne_Rx = 0;     // Décodage de la ligne du bouton sélectionné
      else if (moy_py > 676 && moy_py <= 1125)  antenne_Rx = 1;     //     "
      else if (moy_py > 1126 && moy_py <= 1575) antenne_Rx = 2;
      else if (moy_py > 1576 && moy_py <= 2025) antenne_Rx = 3;
      else if (moy_py > 2026 && moy_py <= 2475) antenne_Rx = 4;
      else if (moy_py > 2476 && moy_py <= 2925) antenne_Rx = 5;
      else if (moy_py > 2926 && moy_py <= 3375) antenne_Rx = 6;
      else if (moy_py > 3376 && moy_py <= 3750) antenne_Rx = 7;
      EEPROM.write(1, antenne_Rx);                                  // Sauvegerder en EEPROM l'antenne RX sélectionnée
      EEPROM.commit();                                              // Sauvegarde finale ici
      vibreur_bip();
    }
    moy_px = 0;                                                     // Ré-initialisation des variables de calcul de coordonnées tactiles moyenne
    moy_py = 0;
    moy_compte = 0;
  }

  // Gestion de clignotement du "TX"
  if ((unsigned long)(millis() - debut_compte_clignot_millis) >= 150)     // Minuterie de clignotement du TX a atteint 150 ms?
  {
    debut_compte_clignot_millis = millis();                               // Oui, ré-initialiser la minuterie de clignotement
    clignot_on = !clignot_on;                                             // Inverser l'état de clignotement
  }

  // Gestion du piézo durant une alarme TX
  if ((alarme_antenne_rx) && (clignot_on)) vibreur_bip();                   // Si une alarme est active et le clignotement est actif, émettre un bip

  // Gestion de l'éteignement de l'affichage
  if ((unsigned long)(millis() - debut_compte_affich_millis) >= 3600000)    // Minuterie d'éteignement de l'affichage a atteint 1 heure?
  {
    digitalWrite(control_eclair_affich, LOW);                               // Oui, Éteindre l'affichage
    debut_compte_affich_millis = millis();                                  // Réinitialiser la minuterie d'éteignement de l'affichage
  }

  // Gestion des antennes lors d'un changement de fréquence sur le transceiver
  if (nouv_frequence_dispo)                                                 // Nouvelle fréquence disponible pour analyse?
  {
    for (int i = 0; i <= 7; i++)                                            // Parcourir les données de toutes les antennes
    {
      if ((frequence >= Data_Antennes[i][2].toInt()) && (frequence < Data_Antennes[i][3].toInt()))    // Trouver l'antenne dont sa plage de fréquence correspond à la fréquence courante
      {
        antenne_Tx = i;                                                     // Sélectionner la bonne antenne une fois trouvée
        antenne_Rx = i;
        nouv_frequence_dispo = false;                                       // Annuler le fanion de nouvelle fréquence disponible
      }
    }
  }
}

// Cette boucle du deuxième coeur monitore la fréquence du transceiver et signale au premier coeur un changement de fréquence
void loop1()
{
  // Gestion des messages CI-V de fréquence courante du transceiver. Signale lorsque les MHz de la fréquence changent
  if (Serial1.available())                      // Octet(s) disponible en tampon du port CI-V?
  {
    rx_data[byte_pos] = Serial1.read();         // Oui, le(s) lire
    //    Serial.println(rx_data[byte_pos]);    // Afficher les octets reçus pour fins de déboggage
    if (rx_data[byte_pos] == 0xFD)              // Octet de fin de message reçu?
    {
      if (( rx_data[0] == 0xfe          // Message de fréquence automatiquement envoyé par le transceiver?...
            && rx_data[1] == 0xfe
            && rx_data[2] == 0x00
            && rx_data[3] == 0x94
            && rx_data[4] == 0x00)
          || ( rx_data[0] == 0xfe         // ...ou message de fréquence suite à une requête manuelle (mise sous tension)?
               && rx_data[1] == 0xfe
               && rx_data[2] == 0xE0
               && rx_data[3] == 0x94
               && rx_data[4] == 0x03))
      {
        frequence = ((rx_data[8] & 0x0F) + (((rx_data[8] >> 4) & 0x0F) * 10));    // Oui, extraire la fréquence (en MHz arrondis) du message
        if (frequence != old_frequence)                                           // Vérifier si c'est une nouvelle fréquence en MHz
        { 
          old_frequence = frequence;                                              // Oui, sauvegarder la nouvelle fréquence comme ancienne fréquence
          nouv_frequence_dispo = true;                                            // Aviser qu'une nouvelle fréquence est disponible
        }
      }
      byte_pos = 0;                                                               // Ré-initialiser le pointeur de position de la chaîne d'octets reçue
    }
    else byte_pos++;                                                              // Caractère de fin de message non reçu: Incrémenter le pointeur de position de la chaîne d'octets reçue
  }
}
// Fin du code source

// Utiliser dans la section de traitement tactile seulement pour trouver les coordonnées tactiles des éléments graphiques durant le développement
//    tft.setTextColor(ILI9341_GREEN);
//    tft.fillRect(0, 0, 100, 50, ILI9341_BLACK);                   // Effacer la nouvelle orientation et les variables
//    tft.setCursor(10, 0);
//    tft.print(moy_px);
//    tft.setCursor(10, 20);
//    tft.print(moy_py);
